package java.util.concurrent;

import java.util.function.DoubleConsumer;
import java.util.function.LongConsumer;
import java.util.function.IntConsumer;
import java.util.Spliterator;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.stream.DoubleStream;
import java.util.stream.LongStream;
import java.util.stream.StreamSupport;
import java.util.stream.IntStream;
import java.security.SecureRandom;
import sun.misc.VM;
import sun.misc.Unsafe;
import java.io.ObjectStreamField;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.Random;

public class ThreadLocalRandom extends Random
{
    private static final AtomicInteger probeGenerator;
    private static final AtomicLong seeder;
    private static final long GAMMA = -7046029254386353131L;
    private static final int PROBE_INCREMENT = -1640531527;
    private static final long SEEDER_INCREMENT = -4942790177534073029L;
    private static final double DOUBLE_UNIT = 1.1102230246251565E-16;
    private static final float FLOAT_UNIT = 5.9604645E-8f;
    private static final ThreadLocal<Double> nextLocalGaussian;
    boolean initialized;
    static final ThreadLocalRandom instance;
    static final String BadBound = "bound must be positive";
    static final String BadRange = "bound must be greater than origin";
    static final String BadSize = "size must be non-negative";
    private static final long serialVersionUID = -5851777807851030925L;
    private static final ObjectStreamField[] serialPersistentFields;
    private static final Unsafe UNSAFE;
    private static final long SEED;
    private static final long PROBE;
    private static final long SECONDARY;
    
    private static long initialSeed() {
        if (Boolean.parseBoolean(VM.getSavedProperty("java.util.secureRandomSeed"))) {
            byte[] seed = SecureRandom.getSeed(8);
            long n = (long)seed[0] & 0xFFL;
            for (int i = 1; i < 8; ++i) {
                n = (n << 8 | ((long)seed[i] & 0xFFL));
            }
            return n;
        }
        return mix64(System.currentTimeMillis()) ^ mix64(System.nanoTime());
    }
    
    private static long mix64(long n) {
        n = (n ^ n >>> 33) * -49064778989728563L;
        n = (n ^ n >>> 33) * -4265267296055464877L;
        return n ^ n >>> 33;
    }
    
    private static int mix32(long n) {
        n = (n ^ n >>> 33) * -49064778989728563L;
        return (int)((n ^ n >>> 33) * -4265267296055464877L >>> 32);
    }
    
    private ThreadLocalRandom() {
        this.initialized = true;
    }
    
    static final void localInit() {
        int addAndGet = ThreadLocalRandom.probeGenerator.addAndGet(-1640531527);
        addAndGet = ((addAndGet == 0) ? 1 : addAndGet);
        long mix64 = mix64(ThreadLocalRandom.seeder.getAndAdd(-4942790177534073029L));
        Thread currentThread = Thread.currentThread();
        ThreadLocalRandom.UNSAFE.putLong(currentThread, ThreadLocalRandom.SEED, mix64);
        ThreadLocalRandom.UNSAFE.putInt(currentThread, ThreadLocalRandom.PROBE, addAndGet);
    }
    
    public static ThreadLocalRandom current() {
        if (ThreadLocalRandom.UNSAFE.getInt(Thread.currentThread(), ThreadLocalRandom.PROBE) == 0) {
            localInit();
        }
        return ThreadLocalRandom.instance;
    }
    
    @Override
    public void setSeed(long n) {
        if (this.initialized) {
            throw new UnsupportedOperationException();
        }
        return;
    }
    
    final long nextSeed() {
        Unsafe unsafe = ThreadLocalRandom.UNSAFE;
        Thread currentThread = Thread.currentThread();
        long n;
        unsafe.putLong(currentThread, ThreadLocalRandom.SEED, n = ThreadLocalRandom.UNSAFE.getLong(currentThread, ThreadLocalRandom.SEED) - 7046029254386353131L);
        return n;
    }
    
    @Override
    protected int next(int n) {
        return (int)(mix64(this.nextSeed()) >>> 64 - n);
    }
    
    final long internalNextLong(long n, long n2) {
        long n3 = mix64(this.nextSeed());
        if (n < n2) {
            long n4 = n2 - n;
            long n5 = n4 - 1L;
            if ((n4 & n5) == 0x0L) {
                n3 = (n3 & n5) + n;
            }
            else if (n4 > 0L) {
                long n7;
                for (long n6 = n3 >>> 1; n6 + n5 - (n7 = n6 % n4) < 0L; n6 = mix64(this.nextSeed()) >>> 1) {}
                n3 = n7 + n;
            }
            else {
                while (n3 < n || n3 >= n2) {
                    n3 = mix64(this.nextSeed());
                }
            }
        }
        return n3;
    }
    
    final int internalNextInt(int n, int n2) {
        int n3 = mix32(this.nextSeed());
        if (n < n2) {
            int n4 = n2 - n;
            int n5 = n4 - 1;
            if ((n4 & n5) == 0x0) {
                n3 = (n3 & n5) + n;
            }
            else if (n4 > 0) {
                int n7;
                for (int n6 = n3 >>> 1; n6 + n5 - (n7 = n6 % n4) < 0; n6 = mix32(this.nextSeed()) >>> 1) {}
                n3 = n7 + n;
            }
            else {
                while (n3 < n || n3 >= n2) {
                    n3 = mix32(this.nextSeed());
                }
            }
        }
        return n3;
    }
    
    final double internalNextDouble(double n, double value) {
        double longBitsToDouble = (double)(this.nextLong() >>> 11) * 1.1102230246251565E-16;
        if (n < value) {
            longBitsToDouble = longBitsToDouble * (value - n) + n;
            if (longBitsToDouble >= value) {
                longBitsToDouble = Double.longBitsToDouble(Double.doubleToLongBits(value) - 1L);
            }
        }
        return longBitsToDouble;
    }
    
    @Override
    public int nextInt() {
        return mix32(this.nextSeed());
    }
    
    @Override
    public int nextInt(int n) {
        if (n <= 0) {
            throw new IllegalArgumentException("bound must be positive");
        }
        int mix32 = mix32(this.nextSeed());
        int n2 = n - 1;
        int n4;
        if ((n & n2) == 0x0) {
            n4 = (mix32 & n2);
        }
        else {
            for (int n3 = mix32 >>> 1; n3 + n2 - (n4 = n3 % n) < 0; n3 = mix32(this.nextSeed()) >>> 1) {}
        }
        return n4;
    }
    
    @Override
    public int nextInt(int n, int n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return this.internalNextInt(n, n2);
    }
    
    @Override
    public long nextLong() {
        return mix64(this.nextSeed());
    }
    
    @Override
    public long nextLong(long n) {
        if (n <= 0L) {
            throw new IllegalArgumentException("bound must be positive");
        }
        long mix64 = mix64(this.nextSeed());
        long n2 = n - 1L;
        long n4;
        if ((n & n2) == 0x0L) {
            n4 = (mix64 & n2);
        }
        else {
            for (long n3 = mix64 >>> 1; n3 + n2 - (n4 = n3 % n) < 0L; n3 = mix64(this.nextSeed()) >>> 1) {}
        }
        return n4;
    }
    
    @Override
    public long nextLong(long n, long n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return this.internalNextLong(n, n2);
    }
    
    @Override
    public double nextDouble() {
        return (double)(mix64(this.nextSeed()) >>> 11) * 1.1102230246251565E-16;
    }
    
    @Override
    public double nextDouble(double value) {
        if (value <= 0.0) {
            throw new IllegalArgumentException("bound must be positive");
        }
        double n = (double)(mix64(this.nextSeed()) >>> 11) * 1.1102230246251565E-16 * value;
        return (n < value) ? n : Double.longBitsToDouble(Double.doubleToLongBits(value) - 1L);
    }
    
    @Override
    public double nextDouble(double n, double n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return this.internalNextDouble(n, n2);
    }
    
    @Override
    public boolean nextBoolean() {
        return mix32(this.nextSeed()) < 0;
    }
    
    @Override
    public float nextFloat() {
        return (float)(mix32(this.nextSeed()) >>> 8) * 5.9604645E-8f;
    }
    
    @Override
    public double nextGaussian() {
        Double n = (Double)ThreadLocalRandom.nextLocalGaussian.get();
        if (n != null) {
            ThreadLocalRandom.nextLocalGaussian.set(null);
            return (double)n;
        }
        double a;
        double n2;
        double n3;
        do {
            n2 = 2.0 * this.nextDouble() - 1.0;
            n3 = 2.0 * this.nextDouble() - 1.0;
            a = n2 * n2 + n3 * n3;
        } while (a >= 1.0 || a == 0.0);
        double sqrt = StrictMath.sqrt(-2.0 * StrictMath.log(a) / a);
        ThreadLocalRandom.nextLocalGaussian.set(new Double(n3 * sqrt));
        return n2 * sqrt;
    }
    
    @Override
    public IntStream ints(long n) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        return StreamSupport.intStream(new ThreadLocalRandom.RandomIntsSpliterator(0L, n, Integer.MAX_VALUE, 0), false);
    }
    
    @Override
    public IntStream ints() {
        return StreamSupport.intStream(new ThreadLocalRandom.RandomIntsSpliterator(0L, Long.MAX_VALUE, Integer.MAX_VALUE, 0), false);
    }
    
    @Override
    public IntStream ints(long n, int n2, int n3) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        if (n2 >= n3) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.intStream(new ThreadLocalRandom.RandomIntsSpliterator(0L, n, n2, n3), false);
    }
    
    @Override
    public IntStream ints(int n, int n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.intStream(new ThreadLocalRandom.RandomIntsSpliterator(0L, Long.MAX_VALUE, n, n2), false);
    }
    
    @Override
    public LongStream longs(long n) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        return StreamSupport.longStream(new ThreadLocalRandom.RandomLongsSpliterator(0L, n, Long.MAX_VALUE, 0L), false);
    }
    
    @Override
    public LongStream longs() {
        return StreamSupport.longStream(new ThreadLocalRandom.RandomLongsSpliterator(0L, Long.MAX_VALUE, Long.MAX_VALUE, 0L), false);
    }
    
    @Override
    public LongStream longs(long n, long n2, long n3) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        if (n2 >= n3) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.longStream(new RandomLongsSpliterator(0L, n, n2, n3), false);
    }
    
    @Override
    public LongStream longs(long n, long n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.longStream(new RandomLongsSpliterator(0L, Long.MAX_VALUE, n, n2), false);
    }
    
    @Override
    public DoubleStream doubles(long n) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        return StreamSupport.doubleStream(new ThreadLocalRandom.RandomDoublesSpliterator(0L, n, Double.MAX_VALUE, 0.0), false);
    }
    
    @Override
    public DoubleStream doubles() {
        return StreamSupport.doubleStream(new ThreadLocalRandom.RandomDoublesSpliterator(0L, Long.MAX_VALUE, Double.MAX_VALUE, 0.0), false);
    }
    
    @Override
    public DoubleStream doubles(long n, double n2, double n3) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        if (n2 >= n3) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.doubleStream(new ThreadLocalRandom.RandomDoublesSpliterator(0L, n, n2, n3), false);
    }
    
    @Override
    public DoubleStream doubles(double n, double n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.doubleStream(new ThreadLocalRandom.RandomDoublesSpliterator(0L, Long.MAX_VALUE, n, n2), false);
    }
    
    static final int getProbe() {
        return ThreadLocalRandom.UNSAFE.getInt(Thread.currentThread(), ThreadLocalRandom.PROBE);
    }
    
    static final int advanceProbe(int x) {
        x ^= x << 13;
        x ^= x >>> 17;
        x ^= x << 5;
        ThreadLocalRandom.UNSAFE.putInt(Thread.currentThread(), ThreadLocalRandom.PROBE, x);
        return x;
    }
    
    static final int nextSecondarySeed() {
        Thread currentThread = Thread.currentThread();
        int int1;
        int x;
        if ((int1 = ThreadLocalRandom.UNSAFE.getInt(currentThread, ThreadLocalRandom.SECONDARY)) != 0) {
            int n = int1 ^ int1 << 13;
            int n2 = n ^ n >>> 17;
            x = (n2 ^ n2 << 5);
        }
        else {
            localInit();
            if ((x = (int)ThreadLocalRandom.UNSAFE.getLong(currentThread, ThreadLocalRandom.SEED)) == 0) {
                x = 1;
            }
        }
        ThreadLocalRandom.UNSAFE.putInt(currentThread, ThreadLocalRandom.SECONDARY, x);
        return x;
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        ObjectOutputStream.PutField putFields = objectOutputStream.putFields();
        putFields.put("rnd", ThreadLocalRandom.UNSAFE.getLong(Thread.currentThread(), ThreadLocalRandom.SEED));
        putFields.put("initialized", true);
        objectOutputStream.writeFields();
    }
    
    private Object readResolve() {
        return current();
    }
    
    static {
        probeGenerator = new AtomicInteger();
        seeder = new AtomicLong(initialSeed());
        nextLocalGaussian = new ThreadLocal();
        instance = new ThreadLocalRandom();
        serialPersistentFields = new ObjectStreamField[] { new ObjectStreamField("rnd", Long.TYPE), new ObjectStreamField("initialized", Boolean.TYPE) };
        try {
            UNSAFE = Unsafe.getUnsafe();
            Class<Object> clazz = Thread.class;
            SEED = ThreadLocalRandom.UNSAFE.objectFieldOffset(clazz.getDeclaredField("threadLocalRandomSeed"));
            PROBE = ThreadLocalRandom.UNSAFE.objectFieldOffset(clazz.getDeclaredField("threadLocalRandomProbe"));
            SECONDARY = ThreadLocalRandom.UNSAFE.objectFieldOffset(clazz.getDeclaredField("threadLocalRandomSecondarySeed"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
    
    static final class RandomIntsSpliterator implements Spliterator.OfInt
    {
        long index;
        final long fence;
        final int origin;
        final int bound;
        
        RandomIntsSpliterator(long index, long fence, int origin, int bound) {
            this.index = index;
            this.fence = fence;
            this.origin = origin;
            this.bound = bound;
        }
        
        @Override
        public ThreadLocalRandom.RandomIntsSpliterator trySplit() {
            long index = this.index;
            long index2 = index + this.fence >>> 1;
            return (index2 <= index) ? null : new ThreadLocalRandom.RandomIntsSpliterator(index, this.index = index2, this.origin, this.bound);
        }
        
        @Override
        public long estimateSize() {
            return this.fence - this.index;
        }
        
        @Override
        public int characteristics() {
            return 17728;
        }
        
        @Override
        public boolean tryAdvance(IntConsumer intConsumer) {
            if (intConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            if (index < this.fence) {
                intConsumer.accept(ThreadLocalRandom.current().internalNextInt(this.origin, this.bound));
                this.index = index + 1L;
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(IntConsumer intConsumer) {
            if (intConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            long fence = this.fence;
            if (index < fence) {
                this.index = fence;
                int origin = this.origin;
                int bound = this.bound;
                ThreadLocalRandom current = ThreadLocalRandom.current();
                do {
                    intConsumer.accept(current.internalNextInt(origin, bound));
                } while (++index < fence);
            }
            return;
        }
    }
    
    static final class RandomLongsSpliterator implements Spliterator.OfLong
    {
        long index;
        final long fence;
        final long origin;
        final long bound;
        
        RandomLongsSpliterator(long index, long fence, long origin, long bound) {
            this.index = index;
            this.fence = fence;
            this.origin = origin;
            this.bound = bound;
        }
        
        @Override
        public ThreadLocalRandom.RandomLongsSpliterator trySplit() {
            long index = this.index;
            long index2 = index + this.fence >>> 1;
            return (index2 <= index) ? null : new ThreadLocalRandom.RandomLongsSpliterator(index, this.index = index2, this.origin, this.bound);
        }
        
        @Override
        public long estimateSize() {
            return this.fence - this.index;
        }
        
        @Override
        public int characteristics() {
            return 17728;
        }
        
        @Override
        public boolean tryAdvance(LongConsumer longConsumer) {
            if (longConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            if (index < this.fence) {
                longConsumer.accept(ThreadLocalRandom.current().internalNextLong(this.origin, this.bound));
                this.index = index + 1L;
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(LongConsumer longConsumer) {
            if (longConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            long fence = this.fence;
            if (index < fence) {
                this.index = fence;
                long origin = this.origin;
                long bound = this.bound;
                ThreadLocalRandom current = ThreadLocalRandom.current();
                do {
                    longConsumer.accept(current.internalNextLong(origin, bound));
                } while (++index < fence);
            }
            return;
        }
    }
    
    static final class RandomDoublesSpliterator implements Spliterator.OfDouble
    {
        long index;
        final long fence;
        final double origin;
        final double bound;
        
        RandomDoublesSpliterator(long index, long fence, double origin, double bound) {
            this.index = index;
            this.fence = fence;
            this.origin = origin;
            this.bound = bound;
        }
        
        @Override
        public ThreadLocalRandom.RandomDoublesSpliterator trySplit() {
            long index = this.index;
            long index2 = index + this.fence >>> 1;
            return (index2 <= index) ? null : new ThreadLocalRandom.RandomDoublesSpliterator(index, this.index = index2, this.origin, this.bound);
        }
        
        @Override
        public long estimateSize() {
            return this.fence - this.index;
        }
        
        @Override
        public int characteristics() {
            return 17728;
        }
        
        @Override
        public boolean tryAdvance(DoubleConsumer doubleConsumer) {
            if (doubleConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            if (index < this.fence) {
                doubleConsumer.accept(ThreadLocalRandom.current().internalNextDouble(this.origin, this.bound));
                this.index = index + 1L;
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(DoubleConsumer doubleConsumer) {
            if (doubleConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            long fence = this.fence;
            if (index < fence) {
                this.index = fence;
                double origin = this.origin;
                double bound = this.bound;
                ThreadLocalRandom current = ThreadLocalRandom.current();
                do {
                    doubleConsumer.accept(current.internalNextDouble(origin, bound));
                } while (++index < fence);
            }
            return;
        }
    }
}
